/* 
   aclsession - ACL session handler	
   Copyright (C) 2003 Max Krasnyanskiy <maxk@qualcomm.com>
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation;
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
   IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY CLAIM,
   OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
   NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
   USE OR PERFORMANCE OF THIS SOFTWARE.
   
   ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS, COPYRIGHTS,
   TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS SOFTWARE IS DISCLAIMED.
 */

/*
 * $Id$
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>
#include <sys/ioctl.h>

#include <sys/socket.h>
#include <sys/uio.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>

static struct hci_dev_info di;

static int sock;
static int device;

static uint16_t ptype = HCI_DH1 | HCI_DH3 | HCI_DH5 | HCI_DM1 | HCI_DM3 | HCI_DM5;
static uint8_t  rswitch = 1;

static uint16_t handle = 0xffff;

int do_open(char *args)
{
	char bdaddr[128];
	bdaddr_t dst;

	if (handle != 0xffff) {
		printf("Session already opened...\n");
		return -1;
	}

	/* Parse arguments */
	sscanf(args, "%s", bdaddr);
	printf("dst: %s\n", bdaddr);

	str2ba(bdaddr, &dst);

	if (hci_create_connection(sock, &dst, ptype, 0, rswitch, &handle, 100000) < 0) {
		printf("Failed to connect to %s. %s(%d).\n", bdaddr, strerror(errno), errno);
		handle = 0xffff;
		return -1;
	}
	
	return 0;
}

void do_close()
{
	if (handle != 0xffff) {
		hci_disconnect(sock, handle, 0x13, 1000);
		handle = -1;
	}
}

int do_sleep(char *args)
{
	int i;

	sscanf(args, "%d", &i);
	printf("Sleep %d seconds\n", i);
	sleep(i);
	return 0;
}

static int get_data(char *str, char *buf)
{
	unsigned char byte[5];
	int len = 0, value;

	while (strlen(str)) {
		sscanf(str, "%s", byte);
		str += strlen(byte) + 1;
		sscanf(byte, "%x", &value);
		buf[len] = value;
		len++;
	}
	return len;
}

int do_send(char *args)
{
	uint8_t data[65535], *ptr;
       	hci_acl_hdr hdr;
	uint8_t  type;
	uint16_t flags;
	struct iovec iv[3];

	int len, i;

	if (handle == 0xffff) {
		printf("Can't send while ACL connection is not opened\n");
		return -1;
	}

	len = get_data(args, data);
	ptr = data;
	
	printf("Sending %d bytes\n", len);
	for (i = 0; i < len; i++)
		printf("0x%2.2X ", data[i]);
	printf("\n");
	
	type = HCI_ACLDATA_PKT;
	iv[0].iov_base = &type;
	iv[0].iov_len  = sizeof(type);
	iv[1].iov_base = &hdr;
	iv[1].iov_len  = sizeof(hdr);
	
	flags = ACL_START;
	while (len) {
		iv[2].iov_base = ptr;
		iv[2].iov_len  = len > di.acl_mtu ? di.acl_mtu : len;

		ptr += iv[2].iov_len;
		len -= iv[2].iov_len;
		
		hdr.handle = htobs(acl_handle_pack(handle, flags));
		hdr.dlen   = htobs(iv[2].iov_len);

		if (writev(sock, iv, 3) < 0) {
			perror("Send failed");
			return -1;
		}

		flags = ACL_CONT;
	}

	return 0;
}

int do_wait(char *args)
{
	uint8_t buf[65535], *ptr;
	struct timeval timeo;
	fd_set fdset;
	int r, i;

	if (handle == 0xffff) {
		printf("Can't wait while ACL connection is not opened\n");
		return -1;
	}

	sscanf(args, "%lu", &timeo.tv_sec);
	timeo.tv_usec = 0;

	printf("Waiting for response\n");

	while (1) {
		hci_acl_hdr *hdr;
		
		FD_ZERO(&fdset);
		FD_SET(sock, &fdset);

		if (!select(sock + 1, &fdset, NULL, NULL, &timeo)) {
			printf("Read timed out\n");
			return -1;
		}

		r = recv(sock, buf, sizeof(buf), 0);
		if (r < 0) {
			perror("Read failed");
			return -1;
		}

		ptr = buf; 

		// Ignore non ACL packets
		// FIXME: Need to check for disconnect event
		if (*ptr != HCI_ACLDATA_PKT)
			continue;
	
		ptr++; r--;
	
		// Check ACL header
		hdr = (void *) ptr;
		ptr += sizeof(hdr); r -= sizeof(hdr);

		if (acl_handle(hdr->handle) != handle)
			continue;
	
		// FIXME: Have to reassemble L2CAP frames here
		
		printf("Read %d bytes\n", r);
		for (i = 0; i < r; i++)
			printf("0x%2.2X ", *ptr++);
		printf("\n");

		break;
	}
	return 0;
}

int handle_input(char *line)
{
	char command[25];

	sscanf(line, "%s", command);
	printf("Got command %s\n", command);
	if (!strncmp(command, "open", 4)) {
		do_open(line + 4);
	} else if (!strncmp(command, "close", 5)) {
		do_close();
	} else if (!strncmp(command, "send", 4)) {
		do_send(line + 4);
	} else if (!strncmp(command, "wait", 4)) {
		do_wait(line + 4);
	} else if (!strncmp(command, "sleep", 5)) {
		do_sleep(line + 5);
	}
	return 0;
}

char *freadline(FILE * f)
{
	static char buffer[65000];
	int c, i = 0;
	int escape_next = 0;
	int comments = 0;

	c = fgetc(f);

	while ((!feof(f)) && (i < 65000)) {
		switch (c) {
		case '#':
			// hash : comments, or not ?
			if (escape_next) {
				buffer[i++] = c;
				escape_next = 0;
				break;
			}
			comments = 1;
			break;

		case '\\':
			// escaping next char
			if (escape_next) {
				buffer[i++] = c;
				escape_next = 0;
				break;
			}
			escape_next = 1;
			break;

		case '\n':
			buffer[i++] = 0;
			return buffer;
			break;

		case EOF:
			// end of file
			buffer[i++] = 0;
			return buffer;
			break;

		default:
			// normal
			if (comments == 0)
				buffer[i++] = c;
			break;
		}
		c = fgetc(f);
	}
	buffer[i++] = 0;
	return buffer;
}

/* -- */
static int open_device(int dev)
{
	struct sockaddr_hci addr;
	struct hci_filter flt;
	int s, opt;

	/* Create HCI socket */
	if ((s=socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI)) < 0) {
		perror("Can't create HCI socket");
		return -1;
	}
	
	opt = 1;
	if (setsockopt(s, SOL_HCI, HCI_DATA_DIR, &opt, sizeof(opt)) < 0) {
		perror("Can't enable data direction info");
		return -1;
	}

	opt = 1;
	if (setsockopt(s, SOL_HCI, HCI_TIME_STAMP, &opt, sizeof(opt)) < 0) {
		perror("Can't enable time stamp");
		return -1;
	}

	/* Setup filter */
	hci_filter_clear(&flt);
	hci_filter_all_ptypes(&flt);
	hci_filter_all_events(&flt);
	if (setsockopt(s, SOL_HCI, HCI_FILTER, &flt, sizeof(flt)) < 0) {
		perror("Can't set HCI filter");
		return -1;
	}

	/* Bind socket to the HCI device */
	addr.hci_family = AF_BLUETOOTH;
	addr.hci_dev = dev;
	if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		printf("Can't attach to device hci%d. %s(%d)\n", 
					dev, strerror(errno), errno);
		return -1;
	}

	/* Set RAW mode */
	if (ioctl(s, HCISETRAW, 1) < 0) {
		perror("Can't set RAW HCI mode");
		return -1;
	}
	
	return s;
}

static void close_device(int s)
{
	/* Set normal mode */
	if (ioctl(s, HCISETRAW, 0) < 0)
		perror("Can't set normal HCI mode");

	close(s);
}
/* -- */

static void usage(void)
{
	printf(
	"Usage: aclsession [OPTION...] [input file]\n"
	"  -i, --device=hci_dev       HCI device\n"
	"  -?, --help                 Give this help list\n"
	"      --usage                Give a short usage message\n"
	);
}

static struct option main_options[] = {
	{"device",	1,0, 'i' },
	{"help", 	0,0, 'h' },
	{ 0 }
};

int main(int argc, char *argv[])
{
	FILE *in;
	char *line;
	int opt;

	printf("ACLSession - ACL session ver %s\n", VERSION);

        while ((opt=getopt_long(argc, argv, "i:xath", main_options, NULL)) != -1) {
                switch(opt) {
		case 'i':
			device = atoi(optarg+3);
			break;

                case 'h':
                default:
                        usage();
                        exit(0);
                }
        }

        argc -= optind;
        argv += optind;
        optind = 0;

	if (argc < 1) {
		usage();
		exit(1);
	}

	sock = open_device(device);
	if (sock < 0) {
		printf("Failed to open device\n");
		return -1;
	}

	di.dev_id = device;
	if (ioctl(sock, HCIGETDEVINFO, (void*) &di) < 0) {
		printf("Can't get device (hci%d) settings. %s(%d)\n", 
					device, strerror(errno), errno);
		close_device(sock);
		return -1;
	}
	
	in = fopen(argv[0], "r");
	if (in == NULL) {
		perror("Failed to open input file");
		exit(1);
	}

	while (!feof(in)) {
		line = freadline(in);
		if (strlen(line))
			handle_input(line);
	}

	close_device(sock);
	return 0;
}
